/* connectsock.c */

/* Bert Bos <bert@w3.org> 12 May 1998 */

static char *version =
"@(#)$Id: connectsock.c,v 1.1 1999/03/08 13:12:36 bbos Exp $";

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <errno.h>
#include <stdlib.h>
#include <string.h>
#include "export.h"
#include "errexit.e"

#ifndef INADDR_NONE
#define INADDR_NONE 0xffffffff
#endif

/* connectsock -- allocate & connect a socket using TCP or UDP */
EXPORT int connectsock(const char *host, const char *service, char *protocol)
{
  /* host = name of host to which connection is desired		*/
  /* service = service associated with the desired port		*/
  /* protocol = name of protocol to use ("tcp" or "udp")	*/
  struct hostent *phe;				/* ptr to host info entry */
  struct servent *pse;				/* ptr to service info entry */
  struct protoent *ppe;				/* ptr protocol info entry */
  struct sockaddr_in sin;			/* Internet endpoint address */
  int s, type;					/* socket desc & socket type */

  memset(&sin, 0, sizeof(sin));
  sin.sin_family = AF_INET;

  /* Map service to port number */
  if ((pse = getservbyname(service, protocol)))
    sin.sin_port = pse->s_port;
  else if ((sin.sin_port = htons(atoi(service))) == 0)
    errexit("can't get \"%s\" service entry\n", service);

  /* Map host name to IP address, allowing for dotted decimal */
  if ((phe = gethostbyname(host)))
    memcpy(&sin.sin_addr, phe->h_addr, phe->h_length);
  else if ((sin.sin_addr.s_addr = inet_addr(host)) == INADDR_NONE)
    errexit("can't get \"%s\" host entry\n", host);

  /* Map protocol name to protocol number */
  if ((ppe = getprotobyname(protocol)) == 0)
    errexit("can't get \"%s\" protocol entry\n", protocol);

  /* Use protocol to choose a socket type */
  if (strcmp(protocol, "udp") == 0)
    type = SOCK_DGRAM;
  else
    type = SOCK_STREAM;

  /* Allocate a socket */
  s = socket(PF_INET, type, ppe->p_proto);
  if (s < 0) errexit("can't create socket: %s\n", strerror(errno));

  /* Connect the socket */
  if (connect(s, (struct sockaddr *)&sin, sizeof(sin)) < 0)
    errexit("can't connect to %s.%s: %s\n", host, service, strerror(errno));

  return s;
}
