%option 8bit caseless nodefault noyywrap
/* %option yylineno */

%{
/* $Id: scan.l,v 1.14 1999/07/31 19:39:06 bbos Exp $ */
#include <assert.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include "export.h"
#include "types.e"
#include "html.h"


EXPORT extern FILE *yyin;
EXPORT int yylineno = 1;


/* esc -- remove outer quotes, escape ", remove \n, return malloc'ed string */
static unsigned char* esc(unsigned char* s)
{
  int i, j;
  unsigned char *u;

  /* Find new length */
  for (i = 0, j = 1; s[j] != s[0]; i++, j++) {
    if (s[j] == '"') i+= 4;
  }
  /* Copy and expand */
  u = malloc(i + 1);
  if (!u) {fprintf(stderr, "Out of memory\n"); exit(1);}
  for (i = 0, j = 1; s[j] != s[0]; i++, j++) {
    if (s[j] == '"')  {strcpy(u + i, "&#34;"); i += 4;}
    else if (s[j] == '\n') u[i] = ' ';		/* \n */
    else if (s[j] == '\r' && s[j+1] == '\n') {u[i] = ' '; j++;}	/* \r\n */
    else if (s[j] == '\r') {u[i] = ' ';}	/* \r */
    else u[i] = s[j];
  }
  u[i] = '\0';
  return u;
}
 
/* strndup -- allocate a string, copy n characters into it and add \0 */
static unsigned char* strndup(const unsigned char *s, size_t n)
{
  unsigned char *t = malloc(n + 1);
  if (!t) {fprintf(stderr, "Out of memory\n"); exit(1);}
  strncpy(t, s, n);
  t[n] = '\0';
  return t;
}

/* lns -- count newlines */
static void lns(const unsigned char *s)
{
  while (*s) {
    if (*s == '\n') yylineno++;
    else if (*s != '\r') ;
    else if (*(s+1) == '\n') {yylineno++; s++;}
    else yylineno++;
    s++;
  }
}

%}

/* thing is rather too permissive, but it will accept <img src=/path>... */

delim		[ \t\r\n\f="'<>]
nondelim	[^ \t\r\n\f="'<>]
name		[a-z0-9:.-]+
thing		{nondelim}+
comment		"<!--"([^-]|-[^-]|--[^>])*"-->"
data		([^<]|<[^a-z/?!])+
doctype		"<!doctype"[ \t\r\n\f]

%s MARKUP VALUE DECL

%%


<INITIAL>"<"{name}	{BEGIN(MARKUP); yylval.s=strdup(yytext+1); return START;}
<INITIAL>"</"({name})?	{BEGIN(MARKUP); yylval.s=strdup(yytext+2); return END;}
<INITIAL>{data}		{yylval.s=strdup(yytext); lns(yytext); return TEXT;}
<INITIAL>{comment}	{yylval.s=strndup(yytext+4,yyleng-7); lns(yytext); return COMMENT;}
<INITIAL>{doctype}	{BEGIN(DECL); lns(yytext+9); return DOCTYPE;}
<INITIAL>"<?"[^>]*">"	{yylval.s=strndup(yytext+2,yyleng-3); lns(yytext); return PROCINS;}

<MARKUP>{name}		{yylval.s = strdup(yytext); return NAME;}
<MARKUP>"="		{BEGIN(VALUE); return '=';}
<MARKUP>[ \t\r\n\f]+	{lns(yytext); /* skip */}
<MARKUP>">"		{BEGIN(INITIAL); return '>';}
<MARKUP>"/>"		{BEGIN(INITIAL); return EMPTYEND;}

<VALUE>[ \t\r\n\f]+	{lns(yytext); /* skip */}
<VALUE>{thing}		{BEGIN(MARKUP); yylval.s=strdup(yytext); return NAME;}
<VALUE>\"[^"]*\"	|
<VALUE>\'[^']*\'	{BEGIN(MARKUP); yylval.s=esc(yytext); lns(yytext); return STRING;}

<DECL>{name}		{yylval.s = strdup(yytext); return NAME;}
<DECL>[ \t\r\n\f]+	{lns(yytext); /* skip */}
<DECL>\"[^"]*\"		|
<DECL>\'[^']*\'		{lns(yytext); yylval.s = esc(yytext); return STRING;}
<DECL>">"		{BEGIN(INITIAL); return '>';}

.			{return *yytext; /* illegal char, in fact */}

%%
/*
 * Local variables:
 * mode: indented-text
 * End:
 */
